#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

/********************************************************/
/*      NCS Algorithms for short byte strings	        */
/*       (result must fit the unsighed int)             */
/* (c) 2018 Sergej Znamenskij <svz@latex.pereslavl.ru>  */
/********************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>

#define PTR_TEST(ptr) { if(! ptr) {\
    fprintf(stderr, "Not memory enough!\n"); return -1;}}

int d_ncs(char *s,char*t);
int d_lcsk(char *s,char*t,int kk);
int d_rouge_w(char *s,char*t);
int d_flcs2(char *s,char*t);
int d_flcs3(char *s,char*t);
int fast_ncs(char *s,char*t);

int d_ncs(char *s,char*t)
{
  unsigned int k,i,j,n,m,x;
  short unsigned int *d;
  n=strlen(s); 
  m=strlen(t);
  d=malloc((sizeof(short unsigned int))*(m+1)*(n+1));
  m++;
  n++;
  for(k=0;k<n;k++)
	d[k]=0;
  for(k=1;k<m;k++)
    d[k*n]=0;
  for(i=1;i<n;i++){
    for(j=1;j<m;j++)
    {
      d[j*n+i]= 
        d[(j-1)*n+i] > d[j*n+i-1]? 
        d[(j-1)*n+i] : d[j*n+i-1];
      for(k=1; k<i+1  &&  k<j+1  &&  s[i-k] == t[j-k] ; k++){
        x=d[(j-k)*n+i-k] + k*(k+1)/2;
        if (x > USHRT_MAX){
          free(d);
          return x;
        }
        if (d[j*n+i] < x) 
          d[j*n+i] = x;
    }}} 
  x=d[n*m-1];
  free(d);
  return x;
}

int d_lcsk(char *s,char*t, int kk)
{
  unsigned int k,i,j,n,m,x;
  short unsigned int *d;
  n=strlen(s); 
  m=strlen(t);
  d=malloc((sizeof(short unsigned int))*(m+1)*(n+1));
  m++;
  n++;
  for(k=0;k<n;k++)
	d[k]=0;
  for(k=1;k<m;k++)
    d[k*n]=0;
  for(i=1;i<n;i++){
    for(j=1;j<m;j++)
    {
      d[j*n+i]= 
        d[(j-1)*n+i] > d[j*n+i-1]? 
        d[(j-1)*n+i] : d[j*n+i-1];
      for(k=kk; k<i+1  &&  k<j+1  &&  s[i-k] == t[j-k] ; k++){
        x=d[(j-k)*n+i-k] + 1;
        if (x > USHRT_MAX){
          free(d);
          return x;
        }
        if (d[j*n+i] < x) 
          d[j*n+i] = x;
    }}} 
  x=d[n*m-1];
  free(d);
  return x;
}

int d_rouge_w(char *s,char*t)
{
  unsigned int k,i,j,n,m,x;
  short unsigned int *d;
  n=strlen(s); 
  m=strlen(t);
  d=malloc((sizeof(short unsigned int))*(m+1)*(n+1));
  m++;
  n++;
  for(k=0;k<n;k++)
	d[k]=0;
  for(k=1;k<m;k++)
    d[k*n]=0;
  for(i=1;i<n;i++){
    for(j=1;j<m;j++)
    {
      d[j*n+i]= 
        d[(j-1)*n+i] > d[j*n+i-1]? 
        d[(j-1)*n+i] : d[j*n+i-1];
      for(k=1; k<i+1  &&  k<j+1  &&  s[i-k] == t[j-k] ; k++){
        x=d[(j-k)*n+i-k] + (int) ( 2*pow(k ,1.4) );
        if (x > USHRT_MAX){
          free(d);
          return x;
        }
        if (d[j*n+i] < x) 
          d[j*n+i] = x;
    }}} 
  x=d[n*m-1];
  free(d);
  return x;
}

int d_flcs2(char *s,char*t)
{
  unsigned int k,i,j,n,m,x;
  short unsigned int *d;
  n=strlen(s); 
  m=strlen(t);
  d=malloc((sizeof(short unsigned int))*(m+1)*(n+1));
  m++;
  n++;
  for(k=0;k<n;k++)
	d[k]=0;
  for(k=1;k<m;k++)
    d[k*n]=0;
  for(i=1;i<n;i++){
    for(j=1;j<m;j++)
    {
      d[j*n+i]= 
        d[(j-1)*n+i] > d[j*n+i-1]? 
        d[(j-1)*n+i] : d[j*n+i-1];
      for(k=1; k<i+1  &&  k<j+1  &&  s[i-k] == t[j-k] ; k++){
        x=d[(j-k)*n+i-k] + k*k;
        if (x > USHRT_MAX){
          free(d);
          return x;
        }
        if (d[j*n+i] < x) 
          d[j*n+i] = x;
    }}} 
  x=d[n*m-1];
  free(d);
  return x;
}

int d_flcs3(char *s,char*t)
{
  unsigned int k,i,j,n,m,x;
  short unsigned int *d;
  n=strlen(s); 
  m=strlen(t);
  d=malloc((sizeof(short unsigned int))*(m+1)*(n+1));
  m++;
  n++;
  for(k=0;k<n;k++)
	d[k]=0;
  for(k=1;k<m;k++)
      d[k*n]=0;
 for(i=1;i<n;i++){
    for(j=1;j<m;j++)
    {
      d[j*n+i]= 
        d[(j-1)*n+i] > d[j*n+i-1]? 
        d[(j-1)*n+i] : d[j*n+i-1];
      for(k=1; k<i+1  &&  k<j+1  &&  s[i-k] == t[j-k] ; k++){
        x=d[(j-k)*n+i-k] + k*k*k;
        if (x > USHRT_MAX){
          free(d);
          return x;
        }
        if (d[j*n+i] < x) 
          d[j*n+i] = x;
    }}} 
  x=d[n*m-1];
  free(d);
  return x;
}


int fast_ncs(char *x,char*y)

{
  long int r; 
  unsigned int i,j,n,m,mi,mx,s,*all0,*all1,*mu,*mp,*pe,*me,*ps,*sw;
  m=strlen(x);
  n=strlen(y); 
  all0=malloc((sizeof (unsigned int))*(3*m+5*(n+1)));
  for( i=0; i < (3*m+5*(n+1)); i++ ){ 
    all0[i]=0;}
  mu   = all0;
  mp   = all0 + n+1;
  pe   = mp + n+1;
  ps   = pe + m+n+1;
  me   = ps + m+n+1;
  for( i=1; i < m+1; i++ ){
    for( j=1; j < n+1; j++ ){
      s = j-i+m;
      mx = mp[j] < mu[j-1]
         ?         mu[j-1]
         : mp[j];
      if( x[i-1] == y[j-1] ){
        if ( ps[s] == 0 ){
           ps[s] = i;  
           me[s] = mp[j-1]+1; 
           if ( mp[j-1] + 1 > mx ){
              pe[s] = i; 
              mu[j] = mp[j-1] + 1; }
           else {
              pe[s] = i+1;   
              mu[j] = mx; }}
        else {
          me[s] += i + 1- ps[s];
          r=mp[j-1] + i - pe[s] +1;
          if (r > UINT_MAX){
                free(all0);
                return UINT_MAX;}
          if ( (me[s] >= mx) && (me[s] >= r)){
               mu[j] = me[s];
               pe[s] = ps[s]; }
          else{
              if ((r  >= mx) && (r >= me[s])){
                  mu[j] = r; }
              else{
                  pe[s] = i+1;   
                  mu[j] = mx; }}}}
      else{ 
          ps[s] = 0;   
          mu[j] = mx; }}
    sw=mu;
    mu=mp;
    mp=sw; } 
  s=mp[n];
  free(all0);
  return s;
}


MODULE = String::BytewiseNCS		PACKAGE = String::BytewiseNCS

int
ncs(s,t)
	char * s
	char * t
CODE:
	RETVAL = fast_ncs(s,t);
OUTPUT:
	RETVAL

int
lcs(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_lcsk(s,t,1);
OUTPUT:
	RETVAL

int
lcs2(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_lcsk(s,t,2);
OUTPUT:
	RETVAL

int
lcs3(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_lcsk(s,t,3);
OUTPUT:
	RETVAL

int
rouge_w(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_rouge_w(s,t);
OUTPUT:
	RETVAL

int
flcs2(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_flcs2(s,t);
OUTPUT:
	RETVAL


int
flcs3(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_flcs3(s,t);
OUTPUT:
	RETVAL

int
raw_ncs(s,t)
	char * s
	char * t
CODE:
	RETVAL = d_ncs(s,t);
OUTPUT:
	RETVAL

#vim:ft=C:ts=8:sw=2:sts=2:set expandtab

